/* $Workfile: DspDemo1.c $       $Revision: 46 $      $Modtime: 2/26/09 2:42p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    DspDemo1.c

    A simple example console application demonstrating the use of the DSP SDK
    example program DspDemo1.out. 
    
    The applications works with the following board combinations:

      Thor-2(8)-PCI / Arni-32-PCI + Vidar-5x4-ASM,

    or OTX PCI Plus series boards,

      Gimle-16-PCI-Plus / Thor-2(4,8)-PCI-Plus + Vidar-5x4-ASM-CST,
      Gimle-16-PCI-Plus / Thor-2(4,8)-PCI-Plus + Vidar-5x4-ASM-PRO,
      Gimle-16-PCI-Plus / Thor-2(4,8)-PCI-Plus + Vidar-5x4-ASM-EX,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express + Vidar-5x4-ASM-CST,
      Thor-2-PCI-Express + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Express + Vidar-5x4-ASM-EX,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or

      Vidar-5x16-PCI

    The example program opens DSP devices and loads the user written DSP SDK
    Example program "DspDemo1.out" to these DSP devices.

    The program utilizes the I/O Control interface to control the blink
    frequency of the C54x heart-beat LED. Simple keystroke commands are
    used to vary the blink frequency.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxDsp.h"

#include <Demo1Io.h>   // IO Control codes for the DSP to host communication
#include <DspIoctl.h>  // IO Control codes for the DSP to host communication

#define DEMO_TASK_REF_DEBUG    501    // Task Reference for a non-blocking call
#define DEMO_DSP_MAX            16    // Max Number of DSP supported
#define DEMO_DEFAULT_LOG_FILE   "DspDemo1.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_FAST_BLINK_COUNT_FACTOR 80
#define DEMO_SLOW_BLINK_COUNT_FACTOR 300

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  enum OtxPhysicalDeviceTypeE nBoardType;
  enum OtxPhysicalDeviceTypeE nAsmBoardType;
  OTX_INT32      nDsps;
  OTX_BOOL       bAsmPresent;
  OTX_BOOL       bUseDspDebugger;
  int            boardNo;
  char           *pszDspProgramName;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
OTX_INT32 DemoProcessUserCommand(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs

static OsaMutexS g_HandleEventCritSec;

struct {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hAsmController0;
  OTX_HANDLE hDsp[DEMO_DSP_MAX];
} g_hPhysDevs;

struct {
  OTX_HANDLE hUserApp[DEMO_DSP_MAX];
} g_hLogicDevs;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Open the Physical Devices on the board
  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;
	OTX_UINT32 nDeviceId;  

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_ARNI_PCI:
      DemoLogf("Opening OTX_DEVICE_ARNI_PCI board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
      DemoLogf("Opening Gimle-16-PCI-Plus board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCIE:
      DemoLogf("Opening Gimle-16-PCIe board\n");
      break;
    case OTX_DEVICE_THOR_PCI:
      DemoLogf("Opening Thor-2-PCI board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening OTX_DEVICE_THOR_2_PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening OTX_DEVICE_THOR_2_PCI-Express board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening OTX_DEVICE_THOR_4_PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI:
      DemoLogf("Opening OTX_DEVICE_THOR_8_PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCMCIA:
      DemoLogf("Opening Thor-2-PCMCIA variant board\n");
      break;
    case OTX_DEVICE_VIDAR_5x_PCI:
      DemoLogf("Opening Vidar-5x16-PCI board\n");
      break;
    default:
      DemoLogf("Opening '%d' board\n", g_demoEnv.nBoardType);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  DemoLogf("Board DeviceId: %d\n", nDeviceId);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    switch (nDeviceId) {
    case 1012:
      DemoLogf("Thor-2-PCMCIA-STD using DspDemo1.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo1.out";
      break;
    case 1043:
      DemoLogf("Thor-2-PCMCIA-PLUS using DspDemo1.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo1.out";
      break;
    case 1057:
      DemoLogf("Thor-2-PCMCIA-CST using DspDemo1Cst.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo1Cst.out";
      break;
    case 1058:
      DemoLogf("Thor-2-PCMCIA-EX using DspDemo1c5410a.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo1c5410a.out";
      break; 
    case 1060:
      DemoLogf("Thor-2-PCMCIA-PRO using DspDemo1c5416.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo1c5416.out";
      break; 
    default:
      DemoLogf("Thor-2-PCMCIA with device id %d is not supported.\n", nDeviceId);
      DemoCheckRc(g_hPhysDevs.hBoardController0, OTX_E_UNEXPECTED);
    }
  }

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_VIDAR_5x_PCI:
  case OTX_DEVICE_THOR_PCMCIA:

    // On Vidar-5x16-PCI and Thor-2-PCMCIA boards, DSPs are owned by the NIC Board Device
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }
    break;
  default:
    {
      // Get a handle to to the ASM Daughter Board
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hBoardController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5x_ASM:
          DemoLogf("Opening Vidar-5x4-ASM board\n");
          bOpenAsm = OTX_TRUE;
          g_demoEnv.pszDspProgramName = "DspDemo1.out";
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_CST:
          DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
          bOpenAsm = OTX_TRUE;
          g_demoEnv.pszDspProgramName = "DspDemo1cst.out";
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_V90:
          DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
          bOpenAsm = OTX_TRUE;
          g_demoEnv.pszDspProgramName = "DspDemo1v90.out";
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_EX:
          DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
          bOpenAsm = OTX_TRUE;
          g_demoEnv.pszDspProgramName = "DspDemo1c5410a.out";
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
          bOpenAsm = OTX_TRUE;
          g_demoEnv.pszDspProgramName = "DspDemo1c5416.out";
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Found Vidar-55x4-ASM board. Not supported by this program. Please see Dsp55Demo1.c instead.\n");
          break;
        }
        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          g_demoEnv.nAsmBoardType  = (enum OtxPhysicalDeviceTypeE)eDevType;
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hAsmController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
        }
      }
      
      for (i = 0; i < g_demoEnv.nDsps; i++) {
        DemoLogf("Opening DSP#%d\n", i);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
      }
    }
    break;

  }

#if 0 // Overriding the Default HPI Buffer configuration (between the driver and the DSP)

  {
    OTX_INT32 nDsp;

    OTX_UINT32 nEventBufStartAddress      = 0x1010;
    OTX_UINT32 nEventBufSize              = 0x200;
    OTX_UINT32 nEventBufProcessingAddress = 0x100F;
    OTX_UINT32 nIoctlStartAddress         = 0x1300;
    OTX_UINT32 nIoctlSize                 = 0x200;
    OTX_BOOL bEventBufIncludeHandle     = OTX_FALSE;
    OTX_BOOL bEventBufIncludeTaskRef    = OTX_FALSE;
    OTX_BOOL bEventBufIncludeTimeStamp  = OTX_FALSE;
    OTX_BOOL bIoctlIncludeHandle        = OTX_TRUE;
    OTX_BOOL bIoctlIncludeTaskRef       = OTX_TRUE;

    for (nDsp = 0; nDsp < g_demoEnv.nDsps; nDsp++) {
      DemoLogf("Default Notification Event Buffer and Io Control attributes settings for DSP#%d:\n", nDsp);
      
      DemoPrintDspAttributes(nDsp);
      
      DemoLogf("Setting new attributes for DSP#%d\n", nDsp);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_ADDR, nEventBufStartAddress); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, nEventBufSize); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_PROCESSING_ADDR, nEventBufProcessingAddress); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_CODE_ADDR, nIoctlStartAddress); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_DATA_MAX_SIZE, nIoctlSize); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_HANDLE, bEventBufIncludeHandle); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TASKREF, bEventBufIncludeTaskRef); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TIMESTAMP, bEventBufIncludeTimeStamp); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_HANDLE, bIoctlIncludeHandle); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_TASKREF, bIoctlIncludeTaskRef); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("New Notification Event Buffer and Io Control attributes settings for DSP#%d:\n", nDsp);
      
      DemoPrintDspAttributes(nDsp);
    }

  }

#endif

  {  
    // Get DSP Highway Speed
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
      DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);
    }


    // Load DSP Program
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      char buf[OTX_MAX_PATH];

      strcpy(buf, g_demoEnv.cwd);
      strcat(buf, g_demoEnv.pszDspProgramName);
      DemoLogf("Loading %s program to DSP#%d\n", buf, i);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], buf, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    
    if (g_demoEnv.bUseDspDebugger) {
      DemoLogf("Please load and run DSP program %s to DSP#%d-%d in your DSP Debugger\n", g_demoEnv.pszDspProgramName, i, g_demoEnv.nDsps-1);
      DemoLogf("Press any key when all the DSP program are running...\n");
      OsaConsoleGetCh();

      // At this point we need to tell the host to restart the programs 
      // (to issue the initial IO commands to the DSP again)
      for (i = 0; i < g_demoEnv.nDsps; i++) {
        DemoLogf("Restarting program on DSP#%d\n", i);
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }
      // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
    }

  }
}


void DemoPrintDspAttributes(OTX_INT32 nDsp)
/*******************************************************************************
  @func
    Queries and prints some of the DSP attributes
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_UINT32 nAttribValueUint32;
  OTX_BOOL nAttribValueBool;
  DemoLogf("Notication Event Buffer Diagnostics for DSP#%d:\n", nDsp);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_ADDR, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf Start Address : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Size (in words)         : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_PROCESSING_ADDR, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Processing Address      : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_CODE_ADDR, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  IoCtl Start Address     : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_DATA_MAX_SIZE, &nAttribValueUint32); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Size (in words)         : 0x%.4X\n", nAttribValueUint32);
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_HANDLE, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf Handle        : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TASKREF, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf TaskRef       : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_INCLUDE_TIMESTAMP, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  Event Buf TimeStamp     : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_HANDLE, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  IoCtl Handle            : %s\n", nAttribValueBool ? "Yes" : "No");
  
  result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_IOCTL_INCLUDE_TASKREF, &nAttribValueBool); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
  DemoLogf("  IoCtl TaskRef           : %s\n", nAttribValueBool ? "Yes" : "No");
  
}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  if (g_demoEnv.bAsmPresent) {
    result = OtxDrvEnable(g_hPhysDevs.hAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
  }

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}


void DemoCreateLogicalDevices(void)
/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    // Create one User Application Logical Devices per DSP
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, i, &g_hLogicDevs.hUserApp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hUserApp[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
  }
}






/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampUs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  // Calculate number of microseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampUs.nUint64 = nTimeStamp.nUint64 / 10 - (nTimeStampSec.nUint64 * 1000000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
#if 0    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
#else
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampUs.sUint32.nLow
      );
#endif
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

void *DemoIdleFunction(
/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  OsaMutexInit(&g_HandleEventCritSec);

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  OsaMutexLock(&g_HandleEventCritSec);

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the DspDemo1.out logical devices
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8192];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8192, &nBytesReturned, &nIoControlCode);
      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        switch(nIoControlCode) {
        case DEMO1_NOTIFY_DEBUG_DATA:
          {
            Demo1NotifyGetDebugDataS *pDataS = (Demo1NotifyGetDebugDataS *)buf;
            if (nBytesReturned != sizeof(Demo1NotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = %.4X\n", pDataS->nData0);
              DemoLogf("nData1 = %.4X\n", pDataS->nData1);
              DemoLogf("nData2 = %.4X\n", pDataS->nData2);

              DemoLogf("sqBuf[299] = %.4X\n", pDataS->sqBuf[299]);
            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }

      }

    } else {

      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }

  OsaMutexUnlock(&g_HandleEventCritSec);


}



static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *pEnv)
/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
{
  short  i;

  // Set reasonable defaults before reading from command line
  pEnv->pszInputStr       = NULL;
  pEnv->bExitOnError      = OTX_FALSE;
  pEnv->nBoardType        = OTX_DEVICE_THOR_PCI;
  pEnv->nAsmBoardType     = OTX_DEVICE_VIDAR_5x_ASM;
  pEnv->nDsps             = 4;
  pEnv->bAsmPresent       = OTX_TRUE;
  pEnv->boardNo           = DEMO_DEFAULT_BOARD_NO;
  pEnv->bUseDspDebugger   = OTX_FALSE;
  pEnv->pszDspProgramName = "DspDemo1.out";

  strcpy(pEnv->logFileName, pEnv->cwd);
  strcat(pEnv->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      pEnv->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(pEnv->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          pEnv->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'N') {   // Board number
          pEnv->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          pEnv->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'A') {
            pEnv->nBoardType = OTX_DEVICE_ARNI_PCI;
          } else if ( *(argv[i]+2) == 'B' || *(argv[i]+2) == '2') {
            pEnv->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if ( *(argv[i]+2) == '4') {
            pEnv->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == '8') {
            pEnv->nBoardType = OTX_DEVICE_THOR_8_PCI;
          } else if ( *(argv[i]+2) == 'P') {
            pEnv->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'T') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'E') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if ( *(argv[i]+2) == 'G') {
            pEnv->nBoardType = OTX_DEVICE_GIMLE_16_PCI_PLUS;
          } else if ( *(argv[i]+2) == '9') {
            pEnv->nBoardType = OTX_DEVICE_GIMLE_16_PCIE;
          } else if ( *(argv[i]+2) == 'V') {
            pEnv->nBoardType  = OTX_DEVICE_VIDAR_5x_PCI;
            pEnv->nDsps       = 16;
            pEnv->bAsmPresent = OTX_FALSE;
          } else if ( *(argv[i]+2) == 'M') {  // Any type of Thor-2-PCMCIA board
            pEnv->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            pEnv->nDsps       = 2;
            pEnv->bAsmPresent = OTX_FALSE;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /N<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|4|8|A|M|G|9|P|T|E|V]: Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-ASM \n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board with Vidar-ASM Plus \n");
  DemoLogf("                  8 = Use Thor-8-PCI Board with Vidar-ASM \n");
  DemoLogf("                  A = Use Arni-16-PCI Board with Vidar-AS M\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA (STD, PLUS, CST, EX, or PRO) Board \n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus Board with Vidar-ASM Plus \n");
  DemoLogf("                  9 = Use Gimle-16-PCIe Board with Vidar-ASM Plus \n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board with Vidar-ASM Plus \n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board with Vidar-ASM Plus \n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board with Vidar-ASM Plus \n");
  DemoLogf("                  V = Use Vidar-5x16-PCI Board \n");
  DemoLogf(" /D             : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  static OTX_INT32      nCurrentUserApp = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" b   Set Blink Count Factor to %d for current DSP User Application \n", DEMO_FAST_BLINK_COUNT_FACTOR);
    DemoLogf(" B   Set Blink Count Factor to %d for current DSP User Application \n", DEMO_SLOW_BLINK_COUNT_FACTOR);
    DemoLogf(" n   Set Count Of Debug Notifications to be sent when 'd' is pressed (default is 1)\n");
    DemoLogf(" e   Enable DSP User Application (Start blinking) on current Dsp\n");
    DemoLogf(" E   Disable DSP User Application (Stop blinking) on current Dsp\n");
    DemoLogf(" d   Get Debug data from the current DSP User Application\n");
    DemoLogf(" D   Get HPI Event Buf Diagnostics data from the current DSP User Application\n");
    DemoLogf(" f   Get Blink Count from the current DSP User Application\n");
    DemoLogf(" r   Reset Blink Count to 0 in current DSP User Application\n");
    DemoLogf(" 0-3 Select active DSP User Application, which is then used for subsequent commands\n");
    DemoLogf(" x   Perform a series of OtxDspIoControl() calls\n");
    DemoLogf(" p   Dump the entire HPI memory for the current DSP\n");
    DemoLogf(" P   Write a value to the last address in the HPI memory for the current DSP\n");
    break;

  case 'b':
    {
      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_FAST_BLINK_COUNT_FACTOR, nCurrentUserApp);

      sIn.nBlinkCountFactor = DEMO_FAST_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 6;
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'x':
    {
      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      static OTX_INT32 nLoopTimes = 1;
      OTX_INT32 i;
      char strBuf[12];
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_FAST_BLINK_COUNT_FACTOR, nCurrentUserApp);
      DemoLogf("How many times? (%d)", nLoopTimes);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nLoopTimes = strtol(strBuf, NULL, 0);
      }

      sIn.nBlinkCountFactor = DEMO_FAST_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 6;

      DemoLogf("Performing %d OtxDspIoControl() calls \n ", nLoopTimes);
      for (i = 0; i < nLoopTimes; i++) {
        result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      }

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'B':
    {
      Demo1IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_SLOW_BLINK_COUNT_FACTOR, nCurrentUserApp);

      sIn.nBlinkCountFactor = DEMO_SLOW_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 7;
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'n':
    {
      Demo1IoInSetCoDebugNotificationsS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo1Rc;
      char szBuf[12];

      DemoLogf("Enter number of Debug Notifications to be sent when 'd' is pressed: ");
      OsaConsoleGetStr(szBuf);
      sIn.coDebugDataNotifications = (unsigned short)strtol(szBuf, NULL, 0);

      DemoLogf("Sending coDebugDataNotifications(%d) to User App #%d ...", sIn.coDebugDataNotifications, nCurrentUserApp);
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_SET_CO_DEBUG_NOTIFICATIONS, &sIn, sizeof(sIn), &nDemo1Rc, sizeof(nDemo1Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo1Rc=%d, nBytesReturned=%d).\n", nDemo1Rc, nBytesReturned);
    }
    break;

  case 'e':
    {
      DemoLogf("Enable Blinking for User App #%d ...", nCurrentUserApp);
      result = OtxDrvEnable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'E':
    {
      DemoLogf("Disable Blinking for User App #%d ...", nCurrentUserApp);
      result = OtxDrvDisable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      DemoLogf("Reset for User App #%d ...", nCurrentUserApp);
      result = OtxDrvReset(g_hLogicDevs.hUserApp[nCurrentUserApp]); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'd':
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for User App #%d (non-blocking call)...", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, DEMO_TASK_REF_DEBUG, 5000); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'f':
    {
      OTX_UINT32 nBytesReturned;
      Demo1IoOutGetBlinkFreqS dataS;

      DemoLogf("Get Blink Frequency for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_BLINK_FREQ, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo1IoOutGetBlinkFreqS)) {
          DemoLogf("Unexpected data length for DEMO1_IOCTL_GET_BLINK_FREQ (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo1 Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("BlinkCountMax = %10d\n", (OTX_UINT32)(dataS.nBlinkCountMaxHigh << 16) + dataS.nBlinkCountMaxLow);
          DemoLogf("BlinkCount    = %10d\n", (OTX_UINT32)(dataS.nBlinkCountHigh << 16) + dataS.nBlinkCountLow);
        }
      }


      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'p':
    {
      OTX_UINT32 i;
      OTX_UINT16 sqReadBuf[0x800];

      DemoLogf("Dump the entire HPI memory for Dsp #%d \n", nCurrentUserApp);

      result = OtxDspReadHpi(g_hPhysDevs.hDsp[nCurrentUserApp], 0x1000, 0x800, sqReadBuf); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      for (i = 0; i < 0x800; i++) {
        if (i && ((i % 8) == 0)) DemoLogf("\n");
        DemoLogf("%.4X ", sqReadBuf[i]);
      }
      DemoLogf("\nDone.\n");
    }
    break;

  case 'P':
    {
      static OTX_UINT16 writeVal = 0x5A5A;
      OTX_UINT16 sqWriteBuf[1];

      sqWriteBuf[0] = writeVal;

      DemoLogf("Write %.4X to the last address in the HPI memory for Dsp #%d \n", writeVal, nCurrentUserApp);

      result = OtxDspWriteHpi(g_hPhysDevs.hDsp[nCurrentUserApp], 0x17FF, 1, sqWriteBuf); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("\nDone.\n");
      writeVal++;
    }
    break;

  case 'D':
    {
      OTX_UINT32 nBytesReturned;
      Demo1IoOutGetHpiEventBufDiagS dataS;

      DemoLogf("Get HPI Event Buf Diagnostics for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO1_IOCTL_GET_HPI_EVENT_BUF_DIAG, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo1IoOutGetHpiEventBufDiagS)) {
          DemoLogf("Unexpected data length for DEMO1_IOCTL_GET_HPI_EVENT_BUF_DIAG (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo1 Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("cwFreeBuffer    = %d\n", dataS.cwFreeBuffer);
          DemoLogf("cwMinFreeBuffer = %d\n", dataS.cwMinFreeBuffer);
          DemoLogf("coLostEvents    = %d\n", dataS.coLostEvents);
        }
      }


      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
    // Select current DSP
    nCurrentUserApp = commandChar - '0';
    DemoLogf("DSP User Application #%d Selected ....\n", nCurrentUserApp);
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command (%d). Press '?' for help, 'q' to quit\n", commandChar);
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("DspDemo1, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DSP SDK Demo Program for OTX PCI/PCMCIA boards.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Connect to the Driver
  DemoLogf("Connecting lib... ");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Query driver revision
  {
    char szVersionInfo[255];
    OtxDrvGetAttributeValueSz(0, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened physical devices...\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened logical devices...\n");
  DemoEnableLogicalDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_INTERNAL); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  DemoLogf("\nMain loop running. Press 'h' for help\n");
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
