/* $Workfile: DspDemo2.c $       $Revision: 51 $      $Modtime: 2/26/09 2:46p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    DspDemo2.c

    A simple example console application demonstrating the use of the DSP SDK
    example program DspDemo2.out. 
    
    The applications works with the following board combinations:

      Thor-2(8)-PCI / Arni-32-PCI + Vidar-5x4-ASM,

    or OTX PCI Plus series boards,

      Gimle-16-PCI-Plus / Thor-2(4,8)-PCI-Plus + Vidar-5x4-ASM-CST,
      Gimle-16-PCI-Plus / Thor-2(4,8)-PCI-Plus + Vidar-5x4-ASM-PRO,
      Gimle-16-PCI-Plus / Thor-2(4,8)-PCI-Plus + Vidar-5x4-ASM-EX,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express + Vidar-5x4-ASM-CST,
      Thor-2-PCI-Express + Vidar-5x4-ASM-PRO,
      Thor-2-PCI-Express + Vidar-5x4-ASM-EX,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA, 
      Thor-2-PCMCIA-Plus, 
      Thor-2-PCMCIA-Ex, 
      Thor-2-PCMCIA-Pro,
      Thor-2-PCMCIA-CST

    or

      Vidar-5x16-PCI

    The example program opens 4 devices and loads the user written DSP SDK
    Example program "DspDemo2.out" to these 4 DSP devices.

    A codec (handset) is connected to each DSP. The DSP code loops the handset
    microphone to the handset speaker. The demo program utilizes the I/O
    Control interface to control speaker gain. Simple keystroke commands are
    used to vary the speaker gain.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#ifdef WIN32 // Testing sound
#include <windows.h>
#include <mmsystem.h>
#endif

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxDsp.h"
#include "OtxT1E1.h"



#include <Demo2Io.h>   // IO Control codes for the DSP to host communication
#include <DspIoctl.h>  // IO Control codes for the DSP to host communication

#define DEMO_TASK_REF_DEBUG 501   // Task Reference for a non-blocking call
#define DEMO_DSP_MAX 4
#define DEMO_NO_OF_LIS 1
#define DEMO_DEFAULT_LOG_FILE "DspDemo2.log"
#define DEMO_RAWSTREAM_FILE   "DspDemo2.raw"
#define DEMO_FULLSTREAM_FILE  "DspDemo2-"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_FAST_BLINK_COUNT_FACTOR 60
#define DEMO_SLOW_BLINK_COUNT_FACTOR 200

typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  enum OtxPhysicalDeviceTypeE nBoardType;
  enum OtxPhysicalDeviceTypeE nAsmBoardType;
  OtxBrdClockSourceE  eClkSource;
  OtxDataFormatE  law;
  OTX_INT32      nDsps;
  OTX_INT32      nT1E1Ts;
  OTX_BOOL       bAsmPresent;
  int            boardNo;
  char           *pszDspProgramName;
  OTX_BOOL       bUseDspDebugger;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
FILE       *g_pDemoStreamFile = NULL;
FILE       *g_pDemoFullStreamFile[DEMO_DSP_MAX];
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bGetSampleThreadsActive = OTX_FALSE;
OTX_UINT32  g_nDspHighwaySpeed;  // DSP Highway speed in kpbs
OTX_BOOL    g_bLoopAllTs[DEMO_DSP_MAX];


struct {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hQuadE1T1[2]; // Only for Thor-8-PCI and Thor-8-PCI-Plus
  OTX_HANDLE hLi[2];
  OTX_HANDLE hTss;
  OTX_HANDLE hAsmController0;
  OTX_HANDLE hDsp[DEMO_DSP_MAX];
  OTX_HANDLE hCodec[4];
} g_hPhysDevs;

struct {
  OTX_HANDLE hUserApp[DEMO_DSP_MAX];
  OTX_INT32 nDev;
} g_hLogicDevs;

struct {
  OTX_HANDLE hFromDsp[4];
  OTX_HANDLE hToDsp[4];
} g_hPipes;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Open the Physical Devices on the Thor-2-PCI board
  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;
	OTX_UINT32 nDeviceId;  

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_ARNI_PCI:
      DemoLogf("Opening OTX_DEVICE_ARNI_PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCI:
      DemoLogf("Opening Thor-2-PCI board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening Thor-2-PCI-Express board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI:
      DemoLogf("Opening Thor-8-PCI board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      DemoLogf("Opening Thor-8-PCI-Plus board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCI_PLUS:
      DemoLogf("Opening Gimle-16-PCI-Plus board\n");
      break;
    case OTX_DEVICE_GIMLE_16_PCIE:
      DemoLogf("Opening Gimle-16-PCIe board\n");
      break;
    case OTX_DEVICE_VIDAR_5x_PCI:
      DemoLogf("Opening Vidar-5x16-PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCMCIA:
      DemoLogf("Opening Thor-2-PCMCIA variant board\n");
      break;
    default:
      DemoLogf("Opening '%d' board\n", g_demoEnv.nBoardType);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < DEMO_NO_OF_LIS; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    {
      OTX_INT32  nQuadE1T1; 
      
      // On the Thor-8-PCI, Gimle-16-PCI-Plus and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open two of those devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < DEMO_NO_OF_LIS; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      OTX_INT32  nQuadE1T1; 
      
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open two of those devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < DEMO_NO_OF_LIS; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    // No T1E1 device on other boards.
    break;
  }


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    // No Codec on this board
    break;
  default:
    // Get a handle to the Time-Space Switch (TSS) on the board
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Enable OTX_DEVICE_TSS\n");
    result = OtxDrvEnable(g_hPhysDevs.hTss,OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hTss, result);
    
    if (g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCI_PLUS && g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCIE) {
      // Get handles to Codec devices #0 - #3
      for (i = 0; i < 4; i++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      }
    }
    break;
  }

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  DemoLogf("Board DeviceId: %d\n", nDeviceId);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    switch (nDeviceId) {
    case 1012:
      DemoLogf("Thor-2-PCMCIA-STD using DspDemo2.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo2.out";
      break;
    case 1043:
      DemoLogf("Thor-2-PCMCIA-PLUS using DspDemo2.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo2.out";
      break;
    case 1057:
      DemoLogf("Thor-2-PCMCIA-CST using DspDemo2Cst.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo2Cst.out";
      break;
    case 1058:
      DemoLogf("Thor-2-PCMCIA-EX using DspDemo2c5410a.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo2c5410a.out";
      break; 
    case 1060:
      DemoLogf("Thor-2-PCMCIA-PRO using DspDemo2c5416.out\n");
      g_demoEnv.pszDspProgramName = "DspDemo2c5416.out";
      break; 
    default:
      DemoLogf("Thor-2-PCMCIA with device id %d is not supported.\n", nDeviceId);
      DemoCheckRc(g_hPhysDevs.hBoardController0, OTX_E_UNEXPECTED);
    }
  }

  if ((g_demoEnv.nBoardType == OTX_DEVICE_VIDAR_5x_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA)) {
    // On Vidar-5x16-PCI and Thor-2-PCMCIA, DSPs are owned by the NIC Board Device
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }

  } else {
    // Get a handle to to the ASM Daughter Board
    OTX_UINT32   eDevType;
    OTX_ITERATOR itrChildTypes;
    OtxDrvIteratorBegin(&itrChildTypes);
    while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hBoardController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
      OTX_BOOL bOpenAsm = OTX_FALSE;
      switch (eDevType) {
      case OTX_DEVICE_VIDAR_5x_ASM:
        DemoLogf("Opening Vidar-5x4-ASM board\n");
        bOpenAsm = OTX_TRUE;
        g_demoEnv.pszDspProgramName = "DspDemo2.out";
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_CST:
        DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
        bOpenAsm = OTX_TRUE;
        g_demoEnv.pszDspProgramName = "DspDemo2cst.out";
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_V90:
        DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
        bOpenAsm = OTX_TRUE;
        g_demoEnv.pszDspProgramName = "DspDemo2v90.out";
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_EX:
        DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
        bOpenAsm = OTX_TRUE;
        g_demoEnv.pszDspProgramName = "DspDemo2c5410a.out";
        break;
      case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
        DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
        bOpenAsm = OTX_TRUE;
        g_demoEnv.pszDspProgramName = "DspDemo2c5416.out";
        break;
      case OTX_DEVICE_VIDAR_55x4_ASM:
        DemoLogf("Found Vidar-55x4-ASM board. Not supported by this program. Please see Dsp55Demo2.c instead.\n");
        break;
      }
      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        g_demoEnv.nAsmBoardType  = (enum OtxPhysicalDeviceTypeE)eDevType;
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hAsmController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      }
    }
    
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
    }

  }

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.nDsps; i++) {
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &g_nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, g_nDspHighwaySpeed);
  }


  // Load DSP Program
  for (i = 0; i < g_demoEnv.nDsps; i++) {
    char buf[OTX_MAX_PATH];

    strcpy(buf, g_demoEnv.cwd);
    strcat(buf, g_demoEnv.pszDspProgramName);
    DemoLogf("Loading %s program to DSP#%d\n", buf, i);
    result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], buf, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  if (g_demoEnv.bUseDspDebugger) {
    int c;
    DemoLogf("************************************************************\n\n");
    DemoLogf("Now please load and run DSP program %s to DSP#%d-%d\n", "Dsp55Demo2.out", 0, g_demoEnv.nDsps-1);
    DemoLogf("in your DSP Debugger.\n\n");
    DemoLogf("After all DSPs are running:\n");
    DemoLogf("  Press 'r' to resynchronize the communication between the DSPs and the host\n");
    DemoLogf("  Press 's' to skip the resynchronization step\n");
    c = OsaConsoleGetCh();
    DemoLogf("\n");

    if (c == 'r' || c == 'R') {
      // At this point we need to tell the host to restart the programs 
      // (to issue the initial IO commands to the DSP again)
      for (i = 0; i < g_demoEnv.nDsps; i++) {
        DemoLogf("Restarting program on DSP#%d\n", i);
        result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      }
      // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
    }
    
  }
}



/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;   // AMI, B8ZS
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;   // AMI, B8ZS
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;    // F12 (D3/4 or SF), F4, ESF, F72 (SLC96)
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;  
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_FALSE;
  pT1CfgData->bfClearChannels            = 0x000000;
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  i = 0;

  // Set termination for the T1/E1 Front End
  result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
  result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
    // T1 mode
    result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
  } else {
    // E1 mode
    result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }
  result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

}




void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  if (g_demoEnv.bAsmPresent) {
    result = OtxDrvEnable(g_hPhysDevs.hAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
  }

  // If board type is Thor-8-PCI, Gimle-16-PCI-Plus or Thor-8-PCI-Plus then enable Quad E1T1 device
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) || (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) || (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE)) {
    OTX_INT32 nQuadE1T1; 

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    OTX_INT32 nQuadE1T1; 

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  if (g_hPhysDevs.hLi[0]) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}



void DemoCreateLogicalDevices(void)
/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    // Create one User Application Logical Devices per DSP
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, i, &g_hLogicDevs.hUserApp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    g_hLogicDevs.nDev = i;
  }
}



void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < g_demoEnv.nDsps; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hUserApp[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
  }
}


void DemoConnectDevices(void)
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.
  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i = 0;
  OTX_UINT8 sqTimeSlotMask[32];

  if (g_demoEnv.nT1E1Ts && g_hPhysDevs.hLi[0]) {
    OTX_INT32 nLi;
    OTX_INT32 nDsp;
    OTX_INT32 nCodec;
    OTX_INT32 nTs;
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      // Connect one E1/T1 timeslot per the DSP (bi-directional)
      // Dsp0 : Connect nT1E1Ts   -> DSP0 -> Codec0 Speaker
      // Dsp1 : Connect nT1E1Ts   <- DSP1 <- Codec0 Microphone
      // Dsp2 : Connect nT1E1Ts+1 -> DSP2 -> Codec1 Speaker
      // Dsp3 : Connect nT1E1Ts+1 <- DSP3 <- Codec1 Microphone
      // Note that this is not a very efficient use of the DSP resources, but
      // it will demonstrate how to make the connections using a DSP 
      // program which is only designed to handle one single timeslot.

      nLi = 0;
      nDsp = i;
      nCodec = i/2;
      nTs = g_demoEnv.nT1E1Ts + i/2;
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[nTs] = 0xFF;
      
      result = OtxDrvCreatePipe(64, &g_hPipes.hFromDsp[i]); DemoCheckRc(g_hPipes.hFromDsp[i], result);
      result = OtxDrvCreatePipe(64, &g_hPipes.hToDsp[i]); DemoCheckRc(g_hPipes.hToDsp[i], result);
      
      if ((i%2) == 0) {
        // Even Li/Dsp number => Li -> Dsp-> Speaker

        DemoLogf("Connecting  Li0Ts%d -> DSP%d -> Codec%d Speaker\n", nTs, nDsp, nCodec); 
        
        // Connect T1/E1 timeslot to DSP.
        result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],   -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hToDsp[i], result);
        result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hToDsp[i], result);
        result = OtxDrvConnectLogicalDevice(g_hPipes.hToDsp[i], -1, g_hLogicDevs.hUserApp[i],   -1); DemoCheckRc(g_hPipes.hToDsp[i], result);
        
        if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA && g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCI_PLUS && g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCIE) {
          // Connect DSP.to Codec speaker
          result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hFromDsp[i], result);
          result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[nCodec],  -1, NULL, 0); DemoCheckRc(g_hPipes.hFromDsp[i], result);
          result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hUserApp[i], -1, g_hPipes.hFromDsp[i], -1); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
        }

      } else {
        // Odd Li/Dsp number => Mic -> Dsp-> Li

        DemoLogf("Connecting  Li0Ts%d <- DSP%d <- Codec%d Mic\n", nTs, nDsp, nCodec); 

        if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA && g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCI_PLUS && g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCIE) {
          // Connect Codec Mic to DSP.
          result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[nCodec], -1, NULL, 0); DemoCheckRc(g_hPipes.hToDsp[i], result);
          result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],     -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hToDsp[i], result);
          result = OtxDrvConnectLogicalDevice(g_hPipes.hToDsp[i], -1, g_hLogicDevs.hUserApp[i],   -1); DemoCheckRc(g_hPipes.hToDsp[i], result);
        }
        
        // Connect DSP.to T1/E1 timeslot
        result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hFromDsp[i], result);
        result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],   -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hFromDsp[i], result);
        result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hUserApp[i], -1, g_hPipes.hFromDsp[i], -1); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);

      }
      
    }


  } else {
    
    // Connect Handsets to the DSPs internally on the board
    // Dsp0 : Codec0 Mic  -> DSP0 -> Codec0 Speaker
    // Dsp1 : Codec1 Mic  -> DSP1 -> Codec1 Speaker
    // Dsp2 : Codec2 Mic  -> DSP2 -> Codec2 Speaker
    // Dsp3 : Codec3 Mic  -> DSP3 -> Codec3 Speaker
    
    for (i = 0; i < g_demoEnv.nDsps; i++) {
      // Create voice pipes for DspX <--> CodecX
      result = OtxDrvCreatePipe(64, &g_hPipes.hFromDsp[i]); DemoCheckRc(g_hPipes.hFromDsp[i], result);
      result = OtxDrvCreatePipe(64, &g_hPipes.hToDsp[i]); DemoCheckRc(g_hPipes.hToDsp[i], result);
      
#if 0 // Letting the OtxDriver pick the timeslot for the DSP highway
      // Connect Codec Microphone to DSP.
      result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hToDsp[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hToDsp[i], result);
      result = OtxDrvConnectLogicalDevice(g_hPipes.hToDsp[i], -1, g_hLogicDevs.hUserApp[i],   -1); DemoCheckRc(g_hPipes.hToDsp[i], result);
      
      // Connect DSP.to Codec Speaker
      result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hFromDsp[i], result);
      result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hFromDsp[i], result);
      result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hUserApp[i], -1, g_hPipes.hFromDsp[i], -1); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
#else // Specific timeslots on the DSP highway
      {
        OTX_UINT8 sqTimeSlotMask[32];
        OTX_INT32 ts = 1;
        
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        
        // Connect Codec Microphone to DSP.
        result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hToDsp[i], result);
        result = OtxDrvConnectPipe(g_hPipes.hToDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hToDsp[i], result);
        result = OtxDrvConnectLogicalDevice(g_hPipes.hToDsp[i], -1, g_hLogicDevs.hUserApp[i],   -1); DemoCheckRc(g_hPipes.hToDsp[i], result);
        
        // Connect DSP.to Codec Speaker
        result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hFromDsp[i], result);
        result = OtxDrvConnectPipe(g_hPipes.hFromDsp[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(g_hPipes.hFromDsp[i], result);
        result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hUserApp[i], -1, g_hPipes.hFromDsp[i], -1); DemoCheckRc(g_hLogicDevs.hUserApp[i], result);
      }
#endif
    }
    
  }


}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}



/*******************************************************************************

  @func

  Simulates Host -> DSP command activity

  @comm

*******************************************************************************/
void *DemoGetSampleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_HANDLE hHandle;
  Demo2IoOutGetSampleS dataS;
  OTX_UINT32 nBytesReturned;

  hHandle = *((OTX_INT32 *) lpParameter);
  
  DemoLogf("[%p]Starting GET sample thread\n", hHandle);

  while (g_bGetSampleThreadsActive) {

    result = OtxDspIoControl(
      hHandle, 
      DEMO2_IOCTL_GET_SAMPLE, 
      NULL, 0, 
      &dataS, sizeof(dataS), 
      &nBytesReturned, 
      OTX_TASK_SYNC, 
      OTX_TIME_INFINITY); 
    DemoCheckRc(hHandle, result);
    
    if (OTX_SUCCESS(result)) {
      if (nBytesReturned != sizeof(Demo2IoOutGetSampleS)) {
        DemoLogf("Unexpected data length for DEMO2_IOCTL_GET_SAMPLE (nBytesReturned=%d)\n", nBytesReturned);
      } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
        DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
      } else {
        //DemoLogf("DataSample = %+.6d\n", (short)dataS.nDataSample);
      }
    }

    //DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    OsaThreadSleep(rand() % 10);

  }
  
  DemoLogf("[%p]DemoGetSampleFunction thread terminating\n", hHandle);

  return 0;

}




void *DemoIdleFunction(
/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];


  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the Demo2.out logical device
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8196];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8196, &nBytesReturned, &nIoControlCode);
      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        switch(nIoControlCode) {
        case DEMO2_NOTIFY_DEBUG_DATA:
          {
            Demo2NotifyGetDebugDataS *pDataS = (Demo2NotifyGetDebugDataS *)buf;

            DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            if (nBytesReturned != sizeof(Demo2NotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              OTX_UINT32 i;
              DemoLogf("DEMO2_NOTIFY_DEBUG_DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = %.4X\n", pDataS->nData0);
              DemoLogf("nData1 = %.4X\n", pDataS->nData1);
              DemoLogf("nData2 = %.4X\n", pDataS->nData2);
              DemoLogf("nData3 = %.4X\n", pDataS->nData3);
              DemoLogf("nData4 = %.4X\n", pDataS->nData4);
              DemoLogf("nData5 = %.4X\n", pDataS->nData5);
              DemoLogf("nData6 = %.4X\n", pDataS->nData6);
              DemoLogf("nData7 = %.4X\n", pDataS->nData7);

              for (i = 0; i < 32; i++) {
                if ((i != 0) && (i%8==0)) DemoLogf("\n");
                DemoLogf("%.2d=%.4X ", i, (unsigned short)pDataS->sqBuf[i]);
              }
              DemoLogf("\n");
            }
          }
          break;
        case DEMO2_NOTIFY_STREAM_RX_DATA:
          {
            Demo2NotifyStreamRxDataS *pDataS = (Demo2NotifyStreamRxDataS *)buf;
            if (nBytesReturned != sizeof(Demo2NotifyStreamRxDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              OTX_INT32 i;
              static OTX_INT32 coReceivedPackets[DEMO_DSP_MAX];
              
              assert(pEventData->m_nSourceId < DEMO_DSP_MAX);

              if (g_pDemoStreamFile) {
                //fwrite(pDataS->nData, 25, sizeof(OTX_INT16), g_pDemoStreamFile);
                for (i = 0; i < 25; i++) {
                  fwrite(&pDataS->nData[i], 1, sizeof(OTX_INT8), g_pDemoStreamFile);
                }
              }

             if ((coReceivedPackets[pEventData->m_nSourceId]++ % 500) == 0) {
                DemoLogf("RX DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
                for (i = 0; i < 25; i++) {
                  DemoLogf("%.4X ", pDataS->nData[i]);
                }
                DemoLogf("\n");

              }

            }
          }
          break;
        case DEMO2_NOTIFY_STREAM_FULL_FRAMES:
          {
            if (nBytesReturned != sizeof(Demo2NotifyStreamFullFramesS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              static OTX_INT32 coReceivedPackets[DEMO_DSP_MAX];
              
              assert(pEventData->m_nSourceId < DEMO_DSP_MAX);
              
              if (g_pDemoFullStreamFile[pEventData->m_nSourceId]) {
                fwrite((OTX_UINT8 *)buf, nBytesReturned, sizeof(OTX_UINT8), g_pDemoFullStreamFile[pEventData->m_nSourceId]);
              }
              
              if ((coReceivedPackets[pEventData->m_nSourceId]++ % 100) == 0) {
                DemoLogf("RX FULL DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              }
              
            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }
      }

    } else {

      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}



static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *pEnv)
/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
{
  short  i;

  // Set reasonable defaults before reading from command line
  pEnv->pszInputStr       = NULL;
  pEnv->bExitOnError      = OTX_FALSE;
  pEnv->nBoardType        = OTX_DEVICE_THOR_PCI;
  pEnv->nAsmBoardType     = OTX_DEVICE_VIDAR_5x_ASM;
  pEnv->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;
  pEnv->nDsps             = 4;
  pEnv->bAsmPresent       = OTX_TRUE;
  pEnv->boardNo           = DEMO_DEFAULT_BOARD_NO;
  pEnv->pszDspProgramName = "DspDemo2.out";
  pEnv->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  pEnv->eT1E1Mode         = OTX_T1E1_MODE_E1;
  pEnv->nT1E1Ts           = 0;
  pEnv->bUseDspDebugger   = OTX_FALSE;

  strcpy(pEnv->logFileName, pEnv->cwd);
  strcat(pEnv->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      pEnv->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(pEnv->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          pEnv->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'N') {   // Board number
          pEnv->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'E') {   // External timeslot
          pEnv->nT1E1Ts = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          pEnv->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'A') {
            pEnv->nBoardType = OTX_DEVICE_ARNI_PCI;
          } else if ( *(argv[i]+2) == '2') {
            pEnv->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if ( *(argv[i]+2) == 'T') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'E') {
            pEnv->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if ( *(argv[i]+2) == '4') {
            pEnv->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == '8') {
            pEnv->nBoardType = OTX_DEVICE_THOR_8_PCI;
          } else if ( *(argv[i]+2) == 'P') {
            pEnv->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'G') {
            pEnv->nBoardType = OTX_DEVICE_GIMLE_16_PCI_PLUS;
          } else if ( *(argv[i]+2) == '9') {
            pEnv->nBoardType = OTX_DEVICE_GIMLE_16_PCIE;
          } else if ( *(argv[i]+2) == 'V') {
            pEnv->nBoardType  = OTX_DEVICE_VIDAR_5x_PCI;
            pEnv->nDsps       = 16;
            pEnv->bAsmPresent = OTX_FALSE;
          } else if ( *(argv[i]+2) == 'M') {  // All Thor-2-PCMCIA models
            pEnv->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            pEnv->nDsps       = 2;
            pEnv->bAsmPresent = OTX_FALSE;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            pEnv->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            pEnv->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            pEnv->eT1E1Mode    = OTX_T1E1_MODE_E1;
            //pEnv->law          = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'T') {   // T1
            pEnv->eT1E1Mode    = OTX_T1E1_MODE_T1;
            //pEnv->law          = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /N<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|4|8|A|G|9|M|P|T|E|V]: Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-ASM\n");
  DemoLogf("                  4 = Use Thor-4-PCI Board with Vidar-ASM Plus\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board with Vidar-ASM\n");
  DemoLogf("                  A = Use Arni-16-PCI Board with Vidar-ASM\n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus Board with Vidar-ASM Plus\n");
  DemoLogf("                  9 = Use Gimle-16-PCIe Board with Vidar-ASM Plus\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA (STD, PLUS, CST, EX, or PRO) Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board with Vidar-ASM Plus\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board with Vidar-ASM Plus\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board with Vidar-ASM Plus\n");
  DemoLogf("                  V = Use Vidar-5x16-PCI Board\n");
  DemoLogf(" /ME            : E1 Li mode (default)\n");
  DemoLogf(" /MT            : T1 Li mode\n");
  DemoLogf(" /KI            : Use Internal board oscillator as clock source\n");
  DemoLogf(" /K0            : Use recovered clock from Line Interface 0 as clock source\n");
  DemoLogf(" /D             : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /E<ts>         : Connect handsets to E1/T1 span (Li0) starting from this timeslot\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}



static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}



OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  static OTX_INT32      nCurrentUserApp = 0;
  static OTX_BOOL       bStreamRxData[DEMO_DSP_MAX];
  static OTX_BOOL       bStreamFullFrames[DEMO_DSP_MAX];
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-3 Select active DSP User Application\n");
    DemoLogf(" g   Set Handset Speaker Gain for current DSP User Application \n");
    DemoLogf(" e   Enable Handset Mic->Speaker Loop on current Dsp\n");
    DemoLogf(" E   Disable Handset Mic->Speaker Loop on current Dsp\n");
    DemoLogf(" d   Get Debug data from the DSP User Application\n");
    DemoLogf(" s   Get Expanded sample value (16-bits) from the Handset Mic\n");
    DemoLogf(" f   Get sample frame (raw %d timeslots) from current Dsp\n", (OTX_INT32)g_nDspHighwaySpeed/64);
    DemoLogf(" t   Enable/Disable continous stream of the received data from Handset Mic\n");
    DemoLogf(" k   Enable/Disable continous stream of full frame to file\n");
    DemoLogf(" j   Start/Stop threads that will continously query the DSP for a sample value\n");
    DemoLogf(" b   Set Blink Count Factor to %d for current DSP User Application \n", DEMO_FAST_BLINK_COUNT_FACTOR);
    DemoLogf(" B   Set Blink Count Factor to %d for current DSP User Application \n", DEMO_SLOW_BLINK_COUNT_FACTOR);
    DemoLogf(" c   Get Frame Buffer Counter value all DSPs\n");
    DemoLogf(" J   Read 16 Bit register (FPGA)\n");
    DemoLogf(" K   Write 16 Bit register (FPGA)\n");
    DemoLogf(" L   Loop all timeslots from in the DSP (override other connections)\n");
    break;

  case 'J':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevs.hBoardController0, nAddress, &readVal); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;

  case 'K':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevs.hBoardController0, nAddress, nValue); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;

  case 'L':
    {
      Demo2IoInLoopAllTsS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      
      g_bLoopAllTs[nCurrentUserApp] = !g_bLoopAllTs[nCurrentUserApp];
      
      DemoLogf("%s looping of all timeslots for User App #%d ...", g_bLoopAllTs[nCurrentUserApp] ? "Enabling" : "Disabling", nCurrentUserApp);
      
      sIn.bLoopAllTs = (OTX_INT16) g_bLoopAllTs[nCurrentUserApp];
      
      result = OtxDspIoControl(
        g_hLogicDevs.hUserApp[nCurrentUserApp], 
        DEMO2_IOCTL_LOOP_ALL_TS, 
        &sIn, sizeof(sIn), 
        &nDemo2Rc, sizeof(nDemo2Rc), 
        &nBytesReturned, 
        OTX_TASK_SYNC, 
        5000 //OTX_TIME_INFINITY
        ); 
      DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      
      DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    }
    break;
  


#if 1 // Just for driver debugging
  case 'y':
    {
      OTX_INT32 i = 0;
      DemoLogf("TSS Test\n");

      for (;;) {
        OtxDrvDbgDump(g_hPhysDevs.hTss, 4);
        DemoLogf("Done (i=%d)\n", i++);
      }

    }
    break;

  case 'Y':
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.nDsps; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;
#endif

#if 1 // Just for driver debugging
  case 'u':
    {
      OTX_INT32 nCapacity = 64;
      OTX_INT32 nLi = 0;
      OTX_INT32 nTs;
      OTX_RESULT result;
      OTX_UINT8 sqTimeSlotMask[128];
      OTX_HANDLE hLiAtoDspPipe[32];
      
      if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) {
      
        // Connect to nLi using timeslot nTs
        // The four Quad-Li highways have 128 timeslots. 
        // TS0 is Li0:TS0, 
        // TS1 is Li1:TS0, 
        // TS2 is Li2:TS0, 
        // TS3 is Li3:TS0, 
        // TS4 is Li0:TS1, 
        // TS5 is Li1:TS1,
        // ..
        // The highway for Quad-Li 0-3 is 0
        // The highway for Quad-Li 4-7 is 1
        // The highway for Quad-Li 8-11 is 2
        // The highway for Quad-Li 12-15 is 3
        // The highway for DSP0 is 4
        // The highway for DSP1 is 5
        // The highway for DSP2 is 6
        // The highway for DSP3 is 7
        
        for (nTs = 2; nTs < 32; nTs++) {
          // Create pipes 
          result = OtxDrvCreatePipe(nCapacity, &hLiAtoDspPipe[nTs]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
          
          memset(sqTimeSlotMask, 0, 128 * sizeof(OTX_UINT8));
          sqTimeSlotMask[nTs*4 + (nLi%4)] = 0xFF;
          result = OtxDrvConnectPipe(hLiAtoDspPipe[nTs], OTX_PIPE_INPUT,  g_hPhysDevs.hTss, 0 + nLi/4, sqTimeSlotMask, 128); DemoCheckRc(hLiAtoDspPipe[nTs], result);
          
          // Connect the output of this pipe to the DSP
          memset(sqTimeSlotMask, 0, 128 * sizeof(OTX_UINT8));
          sqTimeSlotMask[nTs] = 0xFF; 
          result = OtxDrvConnectPipe(hLiAtoDspPipe[nTs], OTX_PIPE_OUTPUT,  g_hPhysDevs.hTss, 4, sqTimeSlotMask, 128); DemoCheckRc(hLiAtoDspPipe[nTs], result);
          
        }
        DemoLogf("Connected pipes for debugging\n");
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {

        // Connect to nLi using timeslot nTs
        // The four Quad-Li highways have 128 timeslots. 
        // TS0 is Li0:TS0, 
        // TS1 is Li1:TS0, 
        // TS2 is Li2:TS0, 
        // TS3 is Li3:TS0, 
        // TS4 is Li0:TS1, 
        // TS5 is Li1:TS1,
        // ..
        // The highway for Quad-Li 0-3 is 0
        // The highway for Quad-Li 4-7 is 1
        // The highway for DSP0 is 8
        // The highway for DSP1 is 9
        // The highway for DSP2 is 10
        // The highway for DSP3 is 11
        
        for (nTs = 2; nTs < 32; nTs++) {
          // Create pipes 
          result = OtxDrvCreatePipe(nCapacity, &hLiAtoDspPipe[nTs]); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
          
          memset(sqTimeSlotMask, 0, 128 * sizeof(OTX_UINT8));
          sqTimeSlotMask[nTs*4 + (nLi%4)] = 0xFF;
          result = OtxDrvConnectPipe(hLiAtoDspPipe[nTs], OTX_PIPE_INPUT,  g_hPhysDevs.hTss, 0 + nLi/4, sqTimeSlotMask, 128); DemoCheckRc(hLiAtoDspPipe[nTs], result);
          
          // Connect the output of this pipe to the DSP
          memset(sqTimeSlotMask, 0, 128 * sizeof(OTX_UINT8));
          sqTimeSlotMask[nTs] = 0xFF; 
          result = OtxDrvConnectPipe(hLiAtoDspPipe[nTs], OTX_PIPE_OUTPUT,  g_hPhysDevs.hTss, 8, sqTimeSlotMask, 32); DemoCheckRc(hLiAtoDspPipe[nTs], result);
          
        }
      } else {
        DemoLogf("This function is not implemented for this board type\n");
      }
    }
    break;

#endif



  case 'c':
    {
      OTX_UINT32 nFrameBufferCounter;
      
      OTX_INT32  i;
      
      for (i = 0; i < g_demoEnv.nDsps; i++) {
        
        result = OtxDrvGetAttributeValueUint32(
          g_hPhysDevs.hDsp[i], 
          OTX_ATTR_DSP_FRAME_BUFFER_COUNTER, 
          &nFrameBufferCounter
          ); 
        DemoCheckRc(g_hPhysDevs.hDsp[i], result);
        
        DemoLogf("DSP[%d]: FrameBufferCounter = %.6d\n", i, nFrameBufferCounter);
      }
      
    }
    break;

  case 'p':
    {
      result = OtxDrvSetAttributeValueBool(
        g_hPhysDevs.hBoardController0, 
        OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
        OTX_TRUE
        ); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      
      DemoLogf("DSP serial port clocks are now disabled\n");
      
    }
    break;

  case 'P':
    {
      result = OtxDrvSetAttributeValueBool(
        g_hPhysDevs.hBoardController0, 
        OTX_ATTR_BRD_DSP_PORTS_DISABLED, 
        OTX_FALSE
        ); 
      DemoCheckRc(g_hPhysDevs.hBoardController0, result);
      
      DemoLogf("DSP serial port clocks are now enabled\n");
      
    }
    break;



  case 'j':
    {
      OTX_INT32  i;
      
      g_bGetSampleThreadsActive = !g_bGetSampleThreadsActive;

      if (g_bGetSampleThreadsActive) {
        for (i = 0; i < g_demoEnv.nDsps; i++) {
          OsaThreadStartEx(DemoGetSampleFunction, (void *) &g_hLogicDevs.hUserApp[i]);
        }
      }
    }
    break;


    
  case 'm':
    {
      OTX_HANDLE hHs3MicPipe;
      OTX_HANDLE hHs2MicPipe;

      // Create voice pipes for Hs2 and Hs3
      result = OtxDrvCreatePipe(64, &hHs3MicPipe); DemoCheckRc(hHs3MicPipe, result);
      result = OtxDrvCreatePipe(64, &hHs2MicPipe); DemoCheckRc(hHs2MicPipe, result);

      // Connect Hs2 <--> Hs3 in the time-space switch
      result = OtxDrvConnectPipe(hHs2MicPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[2], -1, NULL, 0); DemoCheckRc(hHs2MicPipe, result);
      result = OtxDrvConnectPipe(hHs2MicPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[3], -1, NULL, 0); DemoCheckRc(hHs2MicPipe, result);

      result = OtxDrvConnectPipe(hHs3MicPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hCodec[3], -1, NULL, 0); DemoCheckRc(hHs3MicPipe, result);
      result = OtxDrvConnectPipe(hHs3MicPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[2], -1, NULL, 0); DemoCheckRc(hHs3MicPipe, result);
    }
    break;

  case 't':
    {
      Demo2IoInStreamRxDataS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      
      bStreamRxData[nCurrentUserApp] = !bStreamRxData[nCurrentUserApp];
      
      DemoLogf("%s stream of RX data for User App #%d ...", bStreamRxData[nCurrentUserApp] ? "Enabling" : "Disabling", nCurrentUserApp);

#if 1
      if (bStreamRxData[nCurrentUserApp]) {
        if (!g_pDemoStreamFile) {
          g_pDemoStreamFile = fopen(DEMO_RAWSTREAM_FILE, "wb");
          if (!g_pDemoStreamFile) {
            DemoLogf("Error: Could not open file %s for writing\n", DEMO_RAWSTREAM_FILE);
          }
        }
      } else {
        fclose(g_pDemoStreamFile);
        g_pDemoStreamFile = OTX_FALSE;
      }
#endif

      
      sIn.bStreamRxData = (OTX_INT16) bStreamRxData[nCurrentUserApp];
      
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_STREAM_RX_DATA, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      
      DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    }
    break;
  
  case 'k':
    {
      Demo2IoInStreamFullFramesS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      char strBufFileName[128];
      
      bStreamFullFrames[nCurrentUserApp] = !bStreamFullFrames[nCurrentUserApp];
      
      DemoLogf("%s stream of Full Frames for User App #%d ...", bStreamFullFrames[nCurrentUserApp] ? "Enabling" : "Disabling", nCurrentUserApp);

      if (bStreamFullFrames[nCurrentUserApp]) {
        if (!g_pDemoFullStreamFile[nCurrentUserApp]) {
          sprintf(strBufFileName, "%s#%d.dat", DEMO_FULLSTREAM_FILE, nCurrentUserApp);
          g_pDemoFullStreamFile[nCurrentUserApp] = fopen(strBufFileName, "wb");
          if (!g_pDemoFullStreamFile[nCurrentUserApp]) {
            DemoLogf("Error: Could not open file %s for writing\n", strBufFileName);
          } else {
            DemoLogf("Streaming data to file %s for User App #%d\n", strBufFileName, nCurrentUserApp);
          }
        }
      } else {
        fclose(g_pDemoFullStreamFile[nCurrentUserApp]);
        g_pDemoFullStreamFile[nCurrentUserApp] = NULL;
      }
      
      sIn.bStreamFullFrames = (OTX_INT16) bStreamFullFrames[nCurrentUserApp];
      
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_STREAM_FULL_FRAMES, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      
      DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    }
    break;
  
  case 'g':
    {
      Demo2IoInSetTxGainS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      char strBuf[12];

      DemoLogf("Set Handset speaker gain for User App #%d \n", nCurrentUserApp);
      DemoLogf("Please enter gain (-3 <= gain < 7): ");
      OsaConsoleGetStr(strBuf);

      sIn.nTxGain = (OTX_INT16) strtol(strBuf, NULL, 0);

      // The gain range -3 to 7 is arbitrarily range. The DSP code will only use
      // nTxGain to perform an arithmetic shift of the 16-bit value. There is no
      // overflow checking. The range choses in this example is merely a reasonable
      // range for this particular application.
      if ((sIn.nTxGain < -3) || (sIn.nTxGain > 7)) {
        DemoLogf("Error: Gain is out of range\n");
      } else {

        DemoLogf("Setting gain to %+d\n", (OTX_INT16)sIn.nTxGain);
        result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_SET_TX_GAIN, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

        DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
      }
    }
    break;

  case 'b':
    {
      Demo2IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_FAST_BLINK_COUNT_FACTOR, nCurrentUserApp);

      sIn.nBlinkCountFactor = DEMO_FAST_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 6;
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    }
    break;

  case 'B':
    {
      Demo2IoInSetBlinkFreqS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nDemo2Rc;
      DemoLogf("Sending Blink Count factor %d to User App #%d ...", DEMO_SLOW_BLINK_COUNT_FACTOR, nCurrentUserApp);

      sIn.nBlinkCountFactor = DEMO_SLOW_BLINK_COUNT_FACTOR;
      sIn.nDummyData = 7;
      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_SET_BLINK_FREQ, &sIn, sizeof(sIn), &nDemo2Rc, sizeof(nDemo2Rc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nDemo2Rc=%d, nBytesReturned=%d).\n", nDemo2Rc, nBytesReturned);
    }
    break;


  case 'e':
    {
      DemoLogf("Enable Handset loop for User App #%d ...", nCurrentUserApp);
      result = OtxDrvEnable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'E':
    {
      DemoLogf("Disable Handset loop for User App #%d ...", nCurrentUserApp);
      result = OtxDrvDisable(g_hLogicDevs.hUserApp[nCurrentUserApp],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      DemoLogf("Reset Debug Struct User App #%d ...", nCurrentUserApp);
      result = OtxDrvReset(g_hLogicDevs.hUserApp[nCurrentUserApp]); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'd':
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for User App #%d (non-blocking call)...", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, DEMO_TASK_REF_DEBUG, 5000); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 's':
    {
      OTX_UINT32 nBytesReturned;
      Demo2IoOutGetSampleS dataS;

      DemoLogf("Get sample value (16-bit) from the Handset Mic for User App #%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_GET_SAMPLE, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo2IoOutGetSampleS)) {
          DemoLogf("Unexpected data length for DEMO2_IOCTL_GET_SAMPLE (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("DataSample = %+.6d\n", (short)dataS.nDataSample);
        }
      }

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'f':
    {
      OTX_UINT32 nBytesReturned;
      Demo2IoOutGetFullFrameS dataS;

      DemoLogf("\nFrame from Li#%d \n", nCurrentUserApp);

      result = OtxDspIoControl(g_hLogicDevs.hUserApp[nCurrentUserApp], DEMO2_IOCTL_GET_FULL_FRAME, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(Demo2IoOutGetFullFrameS)) {
          DemoLogf("Unexpected data length for DEMO2_IOCTL_GET_FULL_FRAME (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
        } else {
          OTX_INT32 i;
          for (i = 0; i < (OTX_INT32)g_nDspHighwaySpeed/64; i++) {
            if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
            DemoLogf("[%.3d]:%.2X ", i, (OTX_UINT8)dataS.nFrame[i]);
          }
          DemoLogf("\n");

        }
      }

    }
    break;

  case 'x':
    {
      DemoLogf("Disconnecting User App #%d ...", nCurrentUserApp);
      result = OtxDrvDisconnectLogicalDevice(g_hPipes.hToDsp[nCurrentUserApp],    -1, g_hLogicDevs.hUserApp[nCurrentUserApp],   -1); DemoCheckRc(g_hPipes.hToDsp[nCurrentUserApp], result);
      result = OtxDrvDisconnectLogicalDevice(g_hLogicDevs.hUserApp[nCurrentUserApp], -1, g_hPipes.hFromDsp[nCurrentUserApp], -1); DemoCheckRc(g_hLogicDevs.hUserApp[nCurrentUserApp], result);
      DemoLogf("Done.\n");
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
    // Select current DSP
    nCurrentUserApp = commandChar - '0';
    DemoLogf("DSP User Application #%d Selected ....\n", nCurrentUserApp);
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}





int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("DspDemo2, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("DSP SDK Demo Program for OTX PCI boards.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) && (g_demoEnv.nT1E1Ts == 0)) {
    DemoLogf("Error: /E<ts> switch is mandatory for the Thor-2-PCMCIA card, can not = 0\n");
    OsaConsoleExit();
    exit(-1);
  }

  if ((g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) && (g_demoEnv.nT1E1Ts == 0)) {
    DemoLogf("Error: /E<ts> switch is mandatory for the Gimle-16-PCI-Plus card, can not = 0\n");
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  {
    char *pszClockSource;

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      pszClockSource = "Internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
      pszClockSource = "Line Interface 0";
      break;
    default:
      pszClockSource = "Unknown";
      break;
    }

    DemoLogf("Clock Source: %s\n", pszClockSource);
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Configure E1/T1 devices
  if (g_hPhysDevs.hLi[0]) {
    DemoLogf("Configuring E1/T1 devices...\n");
    DemoConfigureE1T1Devices();
  }

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened physical devices...\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened logical devices...\n");
  DemoEnableLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Create pipes and connect the logical devices...\n");
  DemoConnectDevices();

  // Set clock source for the board
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
#if 1 // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.nDsps; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  DemoLogf("\nMain loop running. Press 'h' for help\n");
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", "/E1", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
